@@ -1,6 +1,7 @@
 /*
  * Hydrogen
  * Copyright(c) 2002-2008 by Alex >Comix< Cominu [comix@users.sourceforge.net]
+ * Copyright(c) 2008-2021 The hydrogen development team [hydrogen-devel@lists.sourceforge.net]
  *
  * http://www.hydrogen-music.org
  *
@@ -15,20 +16,19 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program. If not, see https://www.gnu.org/licenses
  *
  */
 
 #include <QtGui>
-#if QT_VERSION >= 0x050000
-#  include <QtWidgets>
-#endif
+#include <QtWidgets>
 #include <QLibraryInfo>
-#include <hydrogen/config.h>
-#include <hydrogen/version.h>
+
+#include <core/config.h>
+#include <core/Version.h>
 #include <getopt.h>
 
+#include "ShotList.h"
 #include "SplashScreen.h"
 #include "HydrogenApp.h"
 #include "MainForm.h"
@@ -36,25 +36,31 @@
 #include "Skin.h"
 
 #ifdef H2CORE_HAVE_LASH
-#include <hydrogen/LashClient.h>
+#include <core/Lash/LashClient.h>
 #endif
 #ifdef H2CORE_HAVE_JACKSESSION
 #include <jack/session.h>
 #endif
 
-#include <hydrogen/midi_map.h>
-#include <hydrogen/audio_engine.h>
-#include <hydrogen/hydrogen.h>
-#include <hydrogen/globals.h>
-#include <hydrogen/event_queue.h>
-#include <hydrogen/Preferences.h>
-#include <hydrogen/h2_exception.h>
-#include <hydrogen/basics/playlist.h>
-#include <hydrogen/helpers/filesystem.h>
-#include <hydrogen/helpers/translations.h>
+#include <core/MidiMap.h>
+#include <core/AudioEngine.h>
+#include <core/Hydrogen.h>
+#include <core/Globals.h>
+#include <core/EventQueue.h>
+#include <core/Preferences.h>
+#include <core/H2Exception.h>
+#include <core/Basics/Playlist.h>
+#include <core/Helpers/Filesystem.h>
+#include <core/Helpers/Translations.h>
+
+#ifdef H2CORE_HAVE_OSC
+#include <core/NsmClient.h>
+#endif
 
 #include <signal.h>
 #include <iostream>
+#include <map>
+#include <set>
 
 //
 // Set the palette used in the application
@@ -65,10 +71,10 @@ void setPalette( QApplication *pQApp )
 	QPalette defaultPalette;
 
 	// A general background color.
-	defaultPalette.setColor( QPalette::Background, QColor( 58, 62, 72 ) );
+	defaultPalette.setColor( QPalette::Window, QColor( 58, 62, 72 ) );
 
 	// A general foreground color.
-	defaultPalette.setColor( QPalette::Foreground, QColor( 255, 255, 255 ) );
+	defaultPalette.setColor( QPalette::WindowText, QColor( 255, 255, 255 ) );
 
 	// Used as the background color for text entry widgets; usually white or another light color.
 	defaultPalette.setColor( QPalette::Base, QColor( 88, 94, 112 ) );
@@ -112,6 +118,20 @@ void setPalette( QApplication *pQApp )
 	pQApp->setStyleSheet("QToolTip {padding: 1px; border: 1px solid rgb(199, 202, 204); background-color: rgb(227, 243, 252); color: rgb(64, 64, 66);}");
 }
 
+// Handle a fatal signal, allowing the logger to complete any outstanding messages before re-raising the
+// signal to allow normal termination.
+static void handleFatalSignal( int nSignal )
+{
+	// First disable signal handler to allow normal termination
+	signal( nSignal, SIG_DFL );
+
+	// Allow logger to complete
+	H2Core::Logger* pLogger = H2Core::Logger::get_instance();
+	if ( pLogger )
+		delete pLogger;
+
+	raise( nSignal );
+}
 
 static int setup_unix_signal_handlers()
 {
@@ -127,8 +147,13 @@ static int setup_unix_signal_handlers()
 		return 1;
 	}
 
-	return 0;
+	for ( int nSignal : { SIGSEGV, SIGILL, SIGFPE, SIGBUS } ) {
+		signal( nSignal, handleFatalSignal );
+	}
+
 #endif
+	
+	return 0;
 }
 
 static void setApplicationIcon(QApplication *app)
@@ -142,6 +167,52 @@ static void setApplicationIcon(QApplication *app)
 	app->setWindowIcon(icon);
 }
 
+
+// QApplication class.
+class H2QApplication : public QApplication {
+
+	QString m_sInitialFileOpen;
+	QWidget *m_pMainForm;
+
+public:
+	H2QApplication( int &argc, char **argv )
+		: QApplication(argc, argv) {
+		m_pMainForm = nullptr;
+	}
+
+	bool event( QEvent *e ) override
+	{
+		if ( e->type() == QEvent::FileOpen ) {
+			QFileOpenEvent *fe = dynamic_cast<QFileOpenEvent*>( e );
+			assert( fe != nullptr );
+
+			if ( m_pMainForm ) {
+				// Forward to MainForm if it's initialised and ready to handle a FileOpenEvent.
+				QApplication::sendEvent( m_pMainForm, e );
+			} else  {
+				// Keep requested file until ready
+				m_sInitialFileOpen = fe->file();
+			}
+			return true;
+		}
+		return QApplication::event( e );
+	}
+
+	// Set the MainForm pointer and forward any requested open event.
+	void setMainForm( QWidget *pMainForm )
+	{
+		m_pMainForm = pMainForm;
+		if ( !m_sInitialFileOpen.isEmpty() ) {
+			QFileOpenEvent ev( m_sInitialFileOpen );
+			QApplication::sendEvent( m_pMainForm, &ev );
+
+		}
+	}
+};
+
+
+
+
 int main(int argc, char *argv[])
 {
 	try {
@@ -151,11 +222,13 @@ int main(int argc, char *argv[])
 
 		// Create bootstrap QApplication to get H2 Core set up with correct Filesystem paths before starting GUI application.
 		QCoreApplication *pBootStrApp = new QCoreApplication( argc, argv );
+		pBootStrApp->setApplicationVersion( QString::fromStdString( H2Core::get_version() ) );
+
 		
 		QCommandLineParser parser;
 		
 		QString aboutText = QString( "\nHydrogen " ) + QString::fromStdString( H2Core::get_version() )  + QString( " [" ) + QString::fromStdString( __DATE__ ) + QString( "]  [http://www.hydrogen-music.org]" ) +
-		QString( "\nCopyright 2002-2008 Alessandro Cominu\nCopyright 2008-2020 The hydrogen development team" ) +
+		QString( "\nCopyright 2002-2008 Alessandro Cominu\nCopyright 2008-2021 The hydrogen development team" ) +
 		QString( "\nHydrogen comes with ABSOLUTELY NO WARRANTY\nThis is free software, and you are welcome to redistribute it under certain conditions. See the file COPYING for details.\n" );
 		
 		parser.setApplicationDescription( aboutText );
@@ -168,6 +241,8 @@ int main(int argc, char *argv[])
 		QCommandLineOption songFileOption( QStringList() << "s" << "song", "Load a song (*.h2song) at startup", "File" );
 		QCommandLineOption kitOption( QStringList() << "k" << "kit", "Load a drumkit at startup", "DrumkitName" );
 		QCommandLineOption verboseOption( QStringList() << "V" << "verbose", "Level, if present, may be None, Error, Warning, Info, Debug or 0xHHHH","Level");
+		QCommandLineOption shotListOption( QStringList() << "t" << "shotlist", "Shot list of widgets to grab", "ShotList" );
+		QCommandLineOption uiLayoutOption( QStringList() << "layout", "UI layout ('tabbed' or 'single')", "Layout" );
 		
 		parser.addHelpOption();
 		parser.addVersionOption();
@@ -179,7 +254,9 @@ int main(int argc, char *argv[])
 		parser.addOption( songFileOption );
 		parser.addOption( kitOption );
 		parser.addOption( verboseOption );
-		
+		parser.addOption( shotListOption );
+		parser.addOption( uiLayoutOption );
+		parser.addPositionalArgument( "file", "Song, playlist or Drumkit file" );
 		
 		//Conditional options
 		#ifdef H2CORE_HAVE_JACKSESSION
@@ -197,6 +274,8 @@ int main(int argc, char *argv[])
 		QString sSongFilename = parser.value ( songFileOption );
 		QString sDrumkitToLoad = parser.value( kitOption );
 		QString sVerbosityString = parser.value( verboseOption );
+		QString sShotList = parser.value( shotListOption );
+		QString sUiLayout = parser.value( uiLayoutOption );
 		
 		unsigned logLevelOpt = H2Core::Logger::Error;
 		if( parser.isSet(verboseOption) ){
@@ -207,6 +286,23 @@ int main(int argc, char *argv[])
 				logLevelOpt = H2Core::Logger::Error|H2Core::Logger::Warning;
 			}
 		}
+
+		// Operating system GUIs typically pass documents to open as
+		// simple positional arguments to the process command
+		// line. Handling this here enables "Open with" as well as
+		// default document bindings to work.
+		QString sArg;
+		foreach ( sArg, parser.positionalArguments() ) {
+			if ( sArg.endsWith( H2Core::Filesystem::songs_ext ) ) {
+				sSongFilename = sArg;
+			}
+			if ( sArg.endsWith( H2Core::Filesystem::drumkit_ext ) ) {
+				sDrumkitName = sArg;
+			}
+			if ( sArg.endsWith( H2Core::Filesystem::playlist_ext ) ) {
+				sPlaylistFilename = sArg;
+			}
+		}
 		
 		#ifdef H2CORE_HAVE_JACKSESSION
 				QString sessionId;
@@ -255,6 +351,15 @@ int main(int argc, char *argv[])
 		QGuiApplication::setHighDpiScaleFactorRoundingPolicy( policy );
 #endif
 
+		// Force layout
+		if ( !sUiLayout.isEmpty() ) {
+			if ( sUiLayout == "tabbed" ) {
+				pPref->setDefaultUILayout( 1 );
+			} else {
+				pPref->setDefaultUILayout( 0 );
+			}
+		}
+
 #ifdef H2CORE_HAVE_LASH
 
 		LashClient::create_instance("hydrogen", "Hydrogen", &argc, &argv);
@@ -270,21 +375,25 @@ int main(int argc, char *argv[])
 			pPref->m_sAudioDriver = "Auto";
 		}
 		else if (sSelectedDriver == "jack") {
-			pPref->m_sAudioDriver = "Jack";
+			pPref->m_sAudioDriver = "JACK";
 		}
 		else if ( sSelectedDriver == "oss" ) {
-			pPref->m_sAudioDriver = "Oss";
+			pPref->m_sAudioDriver = "OSS";
 		}
 		else if ( sSelectedDriver == "alsa" ) {
-			pPref->m_sAudioDriver = "Alsa";
+			pPref->m_sAudioDriver = "ALSA";
 		}
 
 		// Bootstrap is complete, start GUI
 		delete pBootStrApp;
-		QApplication* pQApp = new QApplication( argc, argv );
+		H2QApplication* pQApp = new H2QApplication( argc, argv );
 		pQApp->setApplicationName( "Hydrogen" );
 		pQApp->setApplicationVersion( QString::fromStdString( H2Core::get_version() ) );
 
+		// Process any pending events before showing splash screen. This allows macOS to show previous-crash
+		// warning dialogs before they are covered by the splash screen.
+		pQApp->processEvents();
+
 		QString family = pPref->getApplicationFontFamily();
 		pQApp->setFont( QFont( family, pPref->getApplicationFontPointSize() ) );
 
@@ -329,12 +438,19 @@ int main(int argc, char *argv[])
 
 		SplashScreen *pSplash = new SplashScreen();
 
-		if (bNoSplash) {
+#ifdef H2CORE_HAVE_OSC
+		// Check for being under session management without the
+		// NsmClient class available yet.
+		if ( bNoSplash ||  getenv( "NSM_URL" ) ) {
 			pSplash->hide();
 		}
 		else {
 			pSplash->show();
 		}
+#endif
+#ifndef H2CORE_HAVE_OSC
+		pSplash->show();
+#endif
 
 #ifdef H2CORE_HAVE_LASH
 		if ( H2Core::Preferences::get_instance()->useLash() ){
@@ -373,7 +489,7 @@ int main(int argc, char *argv[])
 			 * here we make it save that hydrogen start in a jacksession case
 			 * every time with jack as audio driver
 			 */
-			pPref->m_sAudioDriver = "Jack";
+			pPref->m_sAudioDriver = "JACK";
 
 		}
 
@@ -392,21 +508,41 @@ int main(int argc, char *argv[])
 		H2Core::Hydrogen::create_instance();
 		
 		// Tell Hydrogen it was started via the QT5 GUI.
-		H2Core::Hydrogen::get_instance()->setActiveGUI( true );
+		H2Core::Hydrogen::get_instance()->setGUIState( H2Core::Hydrogen::GUIState::notReady );
+		
+		// Whether or not to load a default song or supplied one when
+		// constructing the MainForm object.
+		bool bLoadSong = true;
 
-#ifdef H2CORE_HAVE_OSC
 		H2Core::Hydrogen::get_instance()->startNsmClient();
-
-		QString NsmSongFilename = pPref->getNsmSongName();
-
-		if(!NsmSongFilename.isEmpty())
-		{
-			sSongFilename = NsmSongFilename;
+		
+		if ( H2Core::Hydrogen::get_instance()->isUnderSessionManagement() ){
+			
+			// When using the Non Session Management system, the new
+			// Song will be loaded by the NSM client singleton itself
+			// and not by the MainForm. The latter will just access
+			// the already loaded Song.
+			bLoadSong = false;
+			
 		}
-#endif
 
-		MainForm *pMainForm = new MainForm( pQApp, sSongFilename );
+		// If the NSM_URL variable is present, Hydrogen will not
+		// initialize the audio driver and leaves this to the callback
+		// function nsm_open_cb of the NSM client (which will be
+		// called by now). However, the presence of the environmental
+		// variable does not guarantee for a session management and if
+		// no audio driver is initialized yet, we will do it here. 
+		if ( H2Core::Hydrogen::get_instance()->getAudioOutput() == nullptr ) {
+			// Starting drivers can take some time, so show the wait cursor to let the user know that, yes,
+			// we're definitely busy.
+			QApplication::setOverrideCursor( Qt::WaitCursor );
+			H2Core::Hydrogen::get_instance()->restartDrivers();
+			QApplication::restoreOverrideCursor();
+		}
+
+		MainForm *pMainForm = new MainForm( pQApp, sSongFilename, bLoadSong );
 		pMainForm->show();
+		
 		pSplash->finish( pMainForm );
 
 		if( ! sPlaylistFilename.isEmpty() ){
@@ -426,6 +562,22 @@ int main(int argc, char *argv[])
 			} else {
 				___ERRORLOG ( "Error loading the drumkit" );
 			}
+			delete pDrumkitInfo;
+		}
+
+		pQApp->setMainForm( pMainForm );
+
+		// Tell the core that the GUI is now fully loaded and ready.
+		H2Core::Hydrogen::get_instance()->setGUIState( H2Core::Hydrogen::GUIState::ready );
+#ifdef H2CORE_HAVE_OSC
+		if ( NsmClient::get_instance() != nullptr ) {
+			NsmClient::get_instance()->sendDirtyState( false );
+		}
+#endif
+
+		if ( sShotList != QString() ) {
+			ShotList *sl = new ShotList( sShotList );
+			sl->shoot();
 		}
 
 		pQApp->exec();
@@ -441,7 +593,7 @@ int main(int argc, char *argv[])
 		delete MidiActionManager::get_instance();
 
 		___INFOLOG( "Quitting..." );
-		std::cout << "\nBye..." << endl;
+		std::cout << "\nBye..." << std::endl;
 		delete H2Core::Logger::get_instance();
 
 		if (H2Core::Object::count_active()) {
